/*
 * Copyright (C) 2012 SIC! Software GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#define LOG_NDEBUG 0
#define LOG_TAG "SIC! SensorBase.cpp"
#define FUNC_LOG LOGV("%s", __PRETTY_FUNCTION__)

#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <poll.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/select.h>
#include <cutils/log.h>
#include <pthread.h>

#include "SensorBase.h"

//-----------------------------------------------------------------------------
SensorBase::SensorBase(const char* inputName)
    : mEnabled(false)
    , mFirstEventAfterEnableWasSent(false)
    , mLock(PTHREAD_MUTEX_INITIALIZER)
    , mInputEventFd(-EINVAL)
    , mDelay(0)
    , mLastPollTime(0)
    , ipszInputEventName(0)
    , ipszSensorEnableRelativeName(0)
    , ipszSensorSetDelayRelativeFileName(0)
{
    FUNC_LOG;
    mInputEventFd = openInput(inputName);
}

//-----------------------------------------------------------------------------
SensorBase::~SensorBase() {
    FUNC_LOG;
    if (mEnabled) {
        activate( false);
    }
    int err = close(mInputEventFd);
    LOGE_IF((err == -1),"close(mPollFd) errno=%d",errno);
    delete[] ipszInputEventName;
}


//-----------------------------------------------------------------------------
int SensorBase::openSysFsFile(const char* apszDeviceRelativeName)const{
    FUNC_LOG;
    char fullName[PATH_MAX];

//                         "------"       "------------------"
//echo 1 > /sys/class/input/event0/device/vcnl4020/prox_enable
//ipszInputEventName
    strcpy(fullName, "/sys/class/input/");
    strcat(fullName, ipszInputEventName);
    strcat(fullName, "/device/");
    strcat(fullName, apszDeviceRelativeName);
    int fd = open(fullName, O_RDWR);
    if(fd<0){
        LOGE("%s: error open '%s' errno=%d", __PRETTY_FUNCTION__,fullName,errno);
    }
    return fd;
}//SensorBase::openSysFsFile

//-----------------------------------------------------------------------------
//find and open input/event
int SensorBase::openInput(const char* inputName) {
    FUNC_LOG;
    int fd = -1;
    const char *dirname = "/dev/input";
    char devname[PATH_MAX];
    char *filename;
    DIR *dir;
    struct dirent *de;
    dir = opendir(dirname);
    if(dir == NULL)
        return -1;
    strcpy(devname, dirname);
    filename = devname + strlen(devname);
    *filename++ = '/';
    while((de = readdir(dir))) {
        if(de->d_name[0] == '.' &&
                (de->d_name[1] == '\0' ||
                        (de->d_name[1] == '.' && de->d_name[2] == '\0')))
            continue;
        strcpy(filename, de->d_name);
        fd = open(devname, O_RDONLY);
        if (fd>=0) {
            char name[80];
            //get device name
            if (ioctl(fd, EVIOCGNAME(sizeof(name) - 1), &name) < 1) {
                name[0] = '\0';
            }
            if (!strcmp(name, inputName)) {
                delete[] ipszInputEventName;
                ipszInputEventName = new char[strlen(filename) + 1];
                if(ipszInputEventName){
                    strcpy(ipszInputEventName, filename);
                }else{
                    LOGE("%s: unable to allocate memory", __PRETTY_FUNCTION__);
                }
                break;
            } else {
                close(fd);
                fd = -1;
            }
        }
    }
    closedir(dir);
    LOGE_IF(fd<0, "couldn't find '%s' input device", inputName);
    return fd;
}//SensorBase::openInput

//-----------------------------------------------------------------------------
int SensorBase::activate(bool aEnable){
    FUNC_LOG;
    int err = 0;
    pthread_mutex_lock(&mLock);
    if(mEnabled != aEnable) {
        mEnabled = aEnable;
        mFirstEventAfterEnableWasSent=false;
        const int fd = openSysFsFile(ipszSensorEnableRelativeName);
        if (fd < 0) {
            err = -1;
        } else {
            const int len = write(fd, aEnable ? "1" : "0", 2);
            close(fd);
            if(len<0){
               err=len;
            }
        }
    }
    pthread_mutex_unlock(&mLock);
    return err;
}

//-----------------------------------------------------------------------------
int SensorBase::setDelay(int64_t aDelay_ns)
{
    FUNC_LOG;
    int err = 0;
    pthread_mutex_lock(&mLock);
    mDelay = aDelay_ns;
    const int fd = openSysFsFile(ipszSensorSetDelayRelativeFileName);
    if (fd < 0) {
        err = -1;
    } else {
        char buf[21];
        sprintf(buf, "%lld", aDelay_ns);
        const int len = write(fd, buf, strlen(buf)+1);
        close(fd);
        if(len<0){
            err=len;
        }
    }
    pthread_mutex_unlock(&mLock);
    return err;
}

//-----------------------------------------------------------------------------
int SensorBase::poll(sensors_event_t* apData, int count)
{
    FUNC_LOG;
    if (count < 1)
        return -EINVAL;
    return 0;
}
//-----------------------------------------------------------------------------
int64_t SensorBase::getTimestamp() {
//    FUNC_LOG;
    struct timespec t;
    t.tv_sec = t.tv_nsec = 0;
    clock_gettime(CLOCK_MONOTONIC, &t);
    return int64_t(t.tv_sec)*1000000000LL + t.tv_nsec;
}

//-----------------------------------------------------------------------------
int64_t SensorBase::getRemainingPollDelay()const{
    const int64_t KDefaultDelay_ns=500000000LL;//0.5s
    const int64_t d = mLastPollTime+KDefaultDelay_ns-getTimestamp();
    return d;
}

//#############################################################################
//eof SensorBase.cpp

