/*
 * Copyright (C) 2011 The Android Open-Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define LOG_NDEBUG 0
#define LOG_TAG "SIC! sensors.cpp"
#define FUNC_LOG LOGV("%s", __PRETTY_FUNCTION__)
#define FUNC_LOG_END LOGV("//%s", __PRETTY_FUNCTION__)

#include <hardware/sensors.h>
#include <fcntl.h>
#include <errno.h>
#include <dirent.h>
#include <math.h>
#include <poll.h>
#include <pthread.h>
#include <stdlib.h>

#include <linux/input.h>

#include <utils/Atomic.h>
#include <utils/Log.h>

#include "sensors.h"
#include "LightSensor.h"
#include "ProximitySensor.h"


/*****************************************************************************/

//#define SENSORS_ROTATION_VECTOR_HANDLE  (ID_RV)
//#define SENSORS_LINEAR_ACCEL_HANDLE     (ID_LA)
//#define SENSORS_GRAVITY_HANDLE          (ID_GR)
//#define SENSORS_GYROSCOPE_HANDLE        (ID_GY)
//#define SENSORS_ACCELERATION_HANDLE     (ID_A)
//#define SENSORS_MAGNETIC_FIELD_HANDLE   (ID_M)
//#define SENSORS_ORIENTATION_HANDLE      (ID_O)
#define SENSORS_LIGHT_HANDLE            (ID_L)
#define SENSORS_PROXIMITY_HANDLE        (ID_P)
//#define SENSORS_PRESSURE_HANDLE         (ID_PR)

/*****************************************************************************/

/* The SENSORS Module */
#define LOCAL_SENSORS (2)
static struct sensor_t sSensorList[LOCAL_SENSORS] = {
      { "VCNL4020 Light sensor",//name
        "Vishay Intertechnology, Inc.",//vendor
        1,//version
        SENSORS_LIGHT_HANDLE,//handle
        SENSOR_TYPE_LIGHT,//type
        16383.75f,//maxRange //16 bit 0.25lux/bit
        0.25f,//resolution
        0.01f,//power mA (8 measurements / s)
        0,//minDelay
        { }//reserved
      },
      { "VCNL4020 Proximity sensor",
        "Vishay Intertechnology, Inc.",
        1,//version
        SENSORS_PROXIMITY_HANDLE,
        SENSOR_TYPE_PROXIMITY,
        5.0f,//maxRange
        5.0f,//resolution
        0.1f,//power mA (5 measurements / s) IRED 200mA
        0,//minDelay
        { }//reserved
      },
};
static int numSensors = LOCAL_SENSORS;

static int open_sensors(const struct hw_module_t* module, const char* id,
                        struct hw_device_t** device);


static int sensors__get_sensors_list(struct sensors_module_t* module,
                                     struct sensor_t const** list)
{
    FUNC_LOG;
    *list = sSensorList;
    return numSensors;
}

static struct hw_module_methods_t sensors_module_methods = {
        open: open_sensors
};

struct sensors_module_t HAL_MODULE_INFO_SYM = {
        common: {
                tag: HARDWARE_MODULE_TAG,
                version_major: 1,
                version_minor: 0,
                id: SENSORS_HARDWARE_MODULE_ID,
                name: "SIC! Sensor module",
                author: "SIC! Software GmbH",
                methods: &sensors_module_methods,
                dso: 0,
                reserved: {},
        },
        get_sensors_list: sensors__get_sensors_list,
};

struct sensors_poll_context_t {
    struct sensors_poll_device_t device; // must be first

    sensors_poll_context_t();
    ~sensors_poll_context_t();
    int activate(int handle, int enabled);
    int setDelay(int handle, int64_t ns);
    int poll(sensors_event_t* data, int count);
private:
    enum {
        ELight,
        EProximity,
        KNumSensorDrivers,
	KNumPollFds=KNumSensorDrivers
    };
    struct pollfd mPollFds[KNumPollFds];
    SensorBase* mSensors[KNumSensorDrivers];

    int handleToDriver(int handle) const {
//        FUNC_LOG;
	int idx=-EINVAL;
        switch (handle) {
            case ID_L:
                idx=ELight;break;
            case ID_P:
                idx=EProximity;break;
        }
        return idx;
    }
    int64_t mLastLogTime;
    int     mTotalPollCount;
    int     mTotalEventsRead;
};//struct sensors_poll_context_t

/*****************************************************************************/

sensors_poll_context_t::sensors_poll_context_t()
{
    FUNC_LOG;

    numSensors = LOCAL_SENSORS;

    mSensors[ELight] = new LightSensor();
    mPollFds[ELight].fd = mSensors[ELight]->getPollFd();
    mPollFds[ELight].events = POLLIN;
    mPollFds[ELight].revents = 0;

    mSensors[EProximity] = new ProximitySensor();
    mPollFds[EProximity].fd = mSensors[EProximity]->getPollFd();
    mPollFds[EProximity].events = POLLIN;
    mPollFds[EProximity].revents = 0;
    mLastLogTime = 0;
    mTotalPollCount = 0;
    mTotalEventsRead = 0;
    FUNC_LOG_END;
}

sensors_poll_context_t::~sensors_poll_context_t()
{
    FUNC_LOG;
    for (int i=0 ; i<KNumSensorDrivers ; i++) {
        delete mSensors[i];
    }
}

int sensors_poll_context_t::activate(int handle, int enabled)
{
//    FUNC_LOG;
    int index = handleToDriver(handle);
    if (index < 0) return index;
    int err =  mSensors[index]->activate(enabled);
    return err;
}

int sensors_poll_context_t::setDelay(int handle, int64_t ns)
{
//    FUNC_LOG;
    int index = handleToDriver(handle);
    if (index < 0) return index;
    return mSensors[index]->setDelay( ns);
}

//-----------------------------------------------------------------------------
int sensors_poll_context_t::poll(sensors_event_t* pData, int count){
//    FUNC_LOG;
RETRY:
    int totalDataRead = 0;
    int pollResult = 0;
    int j=0;

    do {
        // see if we have some leftover from the last poll()
        for (int i=0 ; count && i<KNumSensorDrivers ; i++) {
	  
            struct pollfd &rPollFd=mPollFds[i];
            SensorBase* const sensor(mSensors[i]);
   
            if ((rPollFd.revents & POLLIN) ){// || (sensor->hasPendingEvents())) {
                const int dataRead = sensor->poll(pData, count);
                if (dataRead < count) {
                    // no more data for this sensor
                    rPollFd.revents = 0;
                }
                count -= dataRead;
                totalDataRead += dataRead;
                pData += dataRead;
            }
        }
        if (count) {
	    
            // we still have some room, so we try to see if we can get
            // some events immediately or just wait if we don't have
            // anything to return

            const int pollTimeout_ms = totalDataRead ? 0 : -1;//100000;//-1 unendlich
            pollResult = ::poll(mPollFds, KNumPollFds, pollTimeout_ms);

            if (pollResult<0) {
                LOGE("poll() failed %d (%s)",errno, strerror(errno));
                return -errno;
            }
        }
        // if we have events and space, go read them
    } while (pollResult && count);
    if(!totalDataRead){
        //we must not return 0
        LOGE("%s: oops !totalDataRead",__PRETTY_FUNCTION__);
        goto RETRY;
    }
    
    mTotalPollCount++;
    mTotalEventsRead+=totalDataRead;
    const int64_t now = SensorBase::getTimestamp();
    const int64_t d = now-mLastLogTime;
    if(d <0ll || d >10300200100ll){//10s
        mLastLogTime=now;
    }
    return totalDataRead;
}

//-----------------------------------------------------------------------------
//c -> c++ wrapper

static int poll__close(struct hw_device_t *dev)
{
//    FUNC_LOG;
    sensors_poll_context_t *ctx = (sensors_poll_context_t *)dev;
    delete ctx;
    return 0;
}

static int poll__activate(struct sensors_poll_device_t *dev,
                          int handle, int enabled)
{
//    FUNC_LOG;
    sensors_poll_context_t *ctx = (sensors_poll_context_t *)dev;
    return ctx->activate(handle, enabled);
}

static int poll__setDelay(struct sensors_poll_device_t *dev,
                          int handle, int64_t ns)
{
//    FUNC_LOG;
    sensors_poll_context_t *ctx = (sensors_poll_context_t *)dev;
    return ctx->setDelay(handle, ns);
}

static int poll__poll(struct sensors_poll_device_t *dev,
                      sensors_event_t* data, int count)
{
//    FUNC_LOG;
    sensors_poll_context_t *ctx = (sensors_poll_context_t *)dev;
    return ctx->poll(data, count);
}

//-----------------------------------------------------------------------------
/** Open a new instance of a sensor device using name */
static int open_sensors(const struct hw_module_t* module, const char* id,
                        struct hw_device_t** device)
{
    FUNC_LOG;if(id){LOGV("id='%s'",id);}
    int status = -EINVAL;
    sensors_poll_context_t *dev = new sensors_poll_context_t();

    memset(&dev->device, 0, sizeof(sensors_poll_device_t));

    dev->device.common.tag = HARDWARE_DEVICE_TAG;
    dev->device.common.version  = 0;
    dev->device.common.module   = const_cast<hw_module_t*>(module);
    dev->device.common.close    = poll__close;
    dev->device.activate        = poll__activate;
    dev->device.setDelay        = poll__setDelay;
    dev->device.poll            = poll__poll;

    *device = &dev->device.common;
    status = 0;

    return status;
}
//#############################################################################
//eof sensors.cpp

