/*
 * board/beagleboard/kernel/drivers/misc/vcnl4020.c - VISHAY  VCNL4020 driver
 *
 * Copyright (C) 2012 SIC! Software GmbH
 *
 *  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * Data sheet at: http://www.vishay.com/docs/83476/vcnl4020.pdf
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/input.h>
#include <linux/err.h>
#include <linux/delay.h>
#include <linux/sysfs.h>
#include <linux/pm_runtime.h>


#define VCNL4020_COMMAND_REGISTER                 (0x80)
#define VCNL4020_COMMAND_REGISTER__ALL_DISABLE    (0x00)
#define VCNL4020_COMMAND_REGISTER__SELFTIMED_EN   (0x01)
#define VCNL4020_COMMAND_REGISTER__PROX_EN        (0x02)
#define VCNL4020_COMMAND_REGISTER__ALS_EN         (0x04)
#define VCNL4020_COMMAND_REGISTER__PROX_OD        (0x08)
#define VCNL4020_COMMAND_REGISTER__ALS_OD         (0x10)
#define VCNL4020_COMMAND_REGISTER__PROX_DATA_RDY  (0x20)
#define VCNL4020_COMMAND_REGISTER__ALS_DATA_RDY   (0x40)
#define VCNL4020_COMMAND_REGISTER__CONFIG_LOCK    (0x80)
#define VCNL4020_COMMAND_REGISTER__WRITE_MASK     (0x1F)

#define VCNL4020_PRODUCT_ID_REVISION_REGISTER     (0x81)
#define VCNL4020_PROXIMITY_RATE_REGISTER          (0x82)
#define VCNL4020_IR_LED_CURRENT_REGISTER          (0x83)
#define VCNL4020_AMBIENT_LIGHT_PARAMETER_REGISTER (0x84)
#define VCNL4020_AMBIENT_LIGHT_RESULT_REGISTER    (0x85)
#define VCNL4020_PROXIMITY_RESULT_REGISTER        (0x87)
#define VCNL4020_INTERRUPT_CONTROL_REGISTER       (0x89)
#define VCNL4020_LOW_THRESHOLD_REGISTER           (0x8A)
#define VCNL4020_HIGH_THRESHOLD_REGISTER          (0x8C)
#define VCNL4020_INTERRUPT_STATUS_REGISTER        (0x8E)
#define VCNL4020_PROXIMITY_MODULAR_TIMING_ADJ     (0x8F)

#define DRIVER_NAME "vcnl4020"

static ktime_t get_range_checked_delay(long long requestdDeleay_ns){
    long long KDefaultDelay=199111000ll;//200ms
    if(requestdDeleay_ns > 10222111000ll){//>10s
        requestdDeleay_ns = KDefaultDelay;
    }else if(requestdDeleay_ns < 50111000ll){//<50ms
        requestdDeleay_ns = KDefaultDelay;
    }
    return ns_to_ktime(requestdDeleay_ns);
}

#define KProxRingBufModuloMask (0x0f)
#define KProxRingBufSize (KProxRingBufModuloMask+1)
#define KValueNear 0
#define KValueFar  5
#define KFixMaxThreshold 60000
#define KFixMinThreshold 3000

static struct MyStatus{
//status
    bool iLuxEnabled;
    bool iProxEnabled;
    bool iLuxRequested;
    bool iProxRequested;
    bool iPowerSuspended;
    int iLuxValue;
    int iProxValue;//the raw 16 bits value from sensor
    int iLuxValueLastEvent;
    int iProxValueLastEvent;
    ktime_t iLuxPollDelay;
    ktime_t iProxPollDelay;
    long long iRequestedLuxPollDelay;
    long long iRequestedProxPollDelay;
    struct workqueue_struct *ipWorkQueue;
    struct work_struct iLux_work;
    struct work_struct iProx_work;
//helper for ProxThreshold
    int iProxValueRound;//KValueNear=0 //KValueFar=5
    int iMaxProx;
    int iMinProx;
    int iProxRingBuf[KProxRingBufSize];
    int iProxRingBufIdx;
    int iProxMaxThreshold;
    int iProxMinThreshold;

//some other helpers
    struct input_dev  *ipProximity_input_dev;
    struct input_dev  *ipLight_input_dev;
    struct i2c_client *ip_i2c_client;
    struct hrtimer iProx_timer;
    struct hrtimer iLux_timer;
    struct mutex iMutex;

//internal logging
    int iLastCommandRegisterRead;
    int iLastCommandRegisterWrite;
//statistic
    int iCountLuxDelayStore;
    int iCountLuxEnableStore;
    int iCountLuxInputShow;
    int iCountLuxTimer;
    int iCountLuxValue;
    int iCountLuxEvent;
    int iCountProxDelayStore;
    int iCountProxEnableStore;
    int iCountProxInputShow;
    int iCountProxTimer;
    int iCountProxValue;
    int iCountProxEvent;
    int iCountStatus;
    int iCountSuspend;
    int iCountResume;
    int iCountInit;
    int iCountExit;
    int iCountProbe;
    int iCountRemove;
    int iCountWriteCommandRegister;
    int iCountBusy;
}s_Status;

//#############################################################################
//interact with sensor chip

//mutex must be hold
static void vcnl4020_read_lux(void){
    int res;
    int newValue;

    //read Hi byte
    res = i2c_smbus_read_byte_data(s_Status.ip_i2c_client, VCNL4020_AMBIENT_LIGHT_RESULT_REGISTER);
    if( res < 0 ) {
        return;
    }
    newValue=res<<8;
    //read Lo Byte
    res = i2c_smbus_read_byte(s_Status.ip_i2c_client);
    if( res < 0 ) {
        return;
    }
    newValue|=res;
    s_Status.iCountLuxValue++;
    s_Status.iLuxValue=newValue;
}
//mutex must be hold
static void vcnl4020_read_prox(void){
    int res;
    int newValue,old;
    bool doCalcThreshold=false;

    //read Hi byte
    res = i2c_smbus_read_byte_data(s_Status.ip_i2c_client, VCNL4020_PROXIMITY_RESULT_REGISTER);
    if( res < 0 ) {
        return;
    }
    newValue=res<<8;
    //read Lo Byte
    res = i2c_smbus_read_byte(s_Status.ip_i2c_client);
    if( res < 0 ) {
        return;
    }
    newValue|=res;
    if(newValue<=0){
        return;
    }

    s_Status.iCountProxValue++;
    s_Status.iProxValue=newValue;

    //store value in ringbuf for avarage
    s_Status.iProxRingBufIdx = (++s_Status.iProxRingBufIdx)&KProxRingBufModuloMask;
    s_Status.iProxRingBuf[s_Status.iProxRingBufIdx]=newValue;

    if(s_Status.iMinProx<=0)s_Status.iMinProx=newValue;
    if(s_Status.iMaxProx<newValue){s_Status.iMaxProx=newValue;doCalcThreshold=true;}
    if(s_Status.iMinProx>newValue || s_Status.iMinProx<=0){s_Status.iMinProx=newValue;doCalcThreshold=true;}

    if(doCalcThreshold){
        int mean=s_Status.iMinProx+100;
        s_Status.iProxMinThreshold=mean;
        s_Status.iProxMaxThreshold=mean+100;
    }

    old=s_Status.iProxValueRound;
    //test fixed thresholds
    if(newValue>KFixMaxThreshold){
        s_Status.iProxValueRound=KValueNear;
        return;
    }
    if(newValue<KFixMinThreshold){
        s_Status.iProxValueRound=KValueFar;
        return;
    }

    //test dynamic threshold
    if(newValue<s_Status.iProxMinThreshold){
        s_Status.iProxValueRound=KValueFar;
        return;
    }
    if(newValue>s_Status.iProxMaxThreshold){
        s_Status.iProxValueRound=KValueNear;
        return;
    }
    //within hysteresis -> leave unchanged
}//vcnl4020_read_prox

static void vcnl4020_update_command_register(void){
    int res;
    u8  val,old;

    //get current status
    do{
        res = i2c_smbus_read_byte_data(s_Status.ip_i2c_client, VCNL4020_COMMAND_REGISTER);
        s_Status.iLastCommandRegisterRead=res;
        if( res < 0 ) {
            res=0;
        }
        //oops there is data -> read them
        if(res&VCNL4020_COMMAND_REGISTER__ALS_DATA_RDY){
            vcnl4020_read_lux();
        }
        if(res&VCNL4020_COMMAND_REGISTER__PROX_DATA_RDY){
            vcnl4020_read_prox();
        }
        //
    }while(   (res&(VCNL4020_COMMAND_REGISTER__ALS_DATA_RDY|VCNL4020_COMMAND_REGISTER__PROX_DATA_RDY))//retry until all data read
          && !(res&(VCNL4020_COMMAND_REGISTER__PROX_EN|VCNL4020_COMMAND_REGISTER__ALS_EN|VCNL4020_COMMAND_REGISTER__SELFTIMED_EN))//but break in selftimed mode
          );

    old = res & VCNL4020_COMMAND_REGISTER__WRITE_MASK;
    if(s_Status.iPowerSuspended){
        val = VCNL4020_COMMAND_REGISTER__ALL_DISABLE;
    }else{
        val = (s_Status.iProxEnabled&s_Status.iProxRequested ? VCNL4020_COMMAND_REGISTER__PROX_OD : 0)  //(0x08)
            | (s_Status.iLuxEnabled &s_Status.iLuxRequested ?  VCNL4020_COMMAND_REGISTER__ALS_OD  : 0)  //(0x10)
              ;
        if(val //we will start new measurement
          && (old &(VCNL4020_COMMAND_REGISTER__PROX_OD|VCNL4020_COMMAND_REGISTER__ALS_OD))//but old one is allready running
         ){//wait for next timer - as we can't restart measuremnt for prox if lux is allready pending and vice versa
            s_Status.iCountBusy++;
            return;
        }
        s_Status.iLuxRequested=false;
        s_Status.iProxRequested=false;
    }
    if(old != val){
        s_Status.iLastCommandRegisterWrite=val;
        res = i2c_smbus_write_byte_data(s_Status.ip_i2c_client, VCNL4020_COMMAND_REGISTER, val);
        if (res < 0) {
        }
        s_Status.iCountWriteCommandRegister++;
    }
}//vcnl4020_update_command_register

//mutex must be hold
//check if result is available
static void vcnl4020_check_result(void){
    int res;

    //test if data ready
    res = i2c_smbus_read_byte_data(s_Status.ip_i2c_client, VCNL4020_COMMAND_REGISTER);
    s_Status.iLastCommandRegisterRead=res;
    if( res >= 0 ) {
        if(res&VCNL4020_COMMAND_REGISTER__ALS_DATA_RDY){
            vcnl4020_read_lux();
        }
        if(res&VCNL4020_COMMAND_REGISTER__PROX_DATA_RDY){
            vcnl4020_read_prox();
        }
    }
    vcnl4020_update_command_register();
}//vcnl4020_check_result

//#############################################################################
//sys fs function

//status for debugging
static ssize_t vcnl4020_driver_status_show(struct device *dev,struct device_attribute *attr, char *buf){
    int maxProxAv=-1;
    int minProxAv=0x7fffFFFF;
    int i;

    s_Status.iCountStatus++;
    for(i=0;i<KProxRingBufSize;i++){
        int t=s_Status.iProxRingBuf[i];
        if(t>maxProxAv)maxProxAv=t;
        if(t<minProxAv)minProxAv=t;
    }

    return sprintf(buf
        ,"+++++++++++VCNL4020 Status v0.3\n"\
         "PowerSuspended=%d\n"\
         "Enabled Lux=%d,%d Prox=%d,%d\n"\
         "Value Lux=%d,%d Prox=%d,%d,%d\n"\
         "PollDelay Lux=%lldns Prox=%lldns\n"\
         "requested Lux=%lldns Prox=%lldns\n"\
         "-----------logging\n"\
         "Max t=%d Av=%d Min t=%d Av=%d\n"\
         "Threshold max=%d min=%d\n"\
         "CommandRegister Write=0x%02x Read=0x%02x\n"\
         "-----------statistic:\n"\
         "Lux  %d,%d,%d,%d,%d,%d\n"\
         "Prox %d,%d,%d,%d,%d,%d\n"\
         "suspend=%d resume=%d init=%d exit=%d probe=%d remove=%d\n"\
         "iCountWriteCommandRegister=%d busy=%d\n"\
         "iCountStatus=%d\n"
         "#####################\n"\
        //------status
        ,s_Status.iPowerSuspended
        ,s_Status.iLuxEnabled,s_Status.iLuxRequested,s_Status.iProxEnabled,s_Status.iProxRequested
        ,s_Status.iLuxValue,s_Status.iLuxValueLastEvent,s_Status.iProxValue,s_Status.iProxValueLastEvent,s_Status.iProxValueRound
        ,ktime_to_ns(s_Status.iLuxPollDelay),ktime_to_ns(s_Status.iProxPollDelay)
        ,s_Status.iRequestedLuxPollDelay,s_Status.iRequestedProxPollDelay
        //------proxy threashhold
        ,s_Status.iMaxProx,maxProxAv,s_Status.iMinProx,minProxAv
        ,s_Status.iProxMaxThreshold,s_Status.iProxMinThreshold
        //------loging
        ,s_Status.iLastCommandRegisterWrite,s_Status.iLastCommandRegisterRead
        //------statistic
        ,s_Status.iCountLuxDelayStore ,s_Status.iCountLuxEnableStore ,s_Status.iCountLuxInputShow ,s_Status.iCountLuxTimer ,s_Status.iCountLuxEvent ,s_Status.iCountLuxValue
        ,s_Status.iCountProxDelayStore,s_Status.iCountProxEnableStore,s_Status.iCountProxInputShow,s_Status.iCountProxTimer,s_Status.iCountProxEvent,s_Status.iCountProxValue
        ,s_Status.iCountSuspend,s_Status.iCountResume,s_Status.iCountInit,s_Status.iCountExit,s_Status.iCountProbe,s_Status.iCountRemove
        ,s_Status.iCountWriteCommandRegister,s_Status.iCountBusy
        ,s_Status.iCountStatus
        );
}//vcnl4020_driver_status_show


//-----------------------------------------------------------------------------
static ssize_t vcnl4020_lux_delay_show(struct device *dev,struct device_attribute *attr, char *buf){
    return sprintf(buf, "%lld\n",ktime_to_ns(s_Status.iLuxPollDelay));
}
//-----------------------------------------------------------------------------
static ssize_t vcnl4020_lux_delay_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t len){
    long long val;

    s_Status.iCountLuxDelayStore++;
    if (strict_strtoll(buf, 10, &val)){
        return -EINVAL;
    }
    mutex_lock(&s_Status.iMutex);
    s_Status.iRequestedLuxPollDelay=val;
    s_Status.iLuxPollDelay = get_range_checked_delay(s_Status.iRequestedLuxPollDelay);
    mutex_unlock(&s_Status.iMutex);
    return len;
}
//-----------------------------------------------------------------------------
static ssize_t vcnl4020_lux_enable_show(struct device *dev,struct device_attribute *attr, char *buf){
    return sprintf(buf, "%d\n",s_Status.iLuxEnabled);
}
//-----------------------------------------------------------------------------
static ssize_t vcnl4020_lux_enable_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t len){
    bool new_value;
    bool valueHasChanged;

    s_Status.iCountLuxEnableStore++;
    if (sysfs_streq(buf, "1")){
        new_value = true;
    } else if (sysfs_streq(buf, "0")){
        new_value = false;
    } else {
        return -EINVAL;
    }

    mutex_lock(&s_Status.iMutex);
    valueHasChanged = s_Status.iLuxEnabled != new_value;
    if(valueHasChanged){
        s_Status.iLuxEnabled = new_value;
        s_Status.iLuxValueLastEvent=-1;
        vcnl4020_update_command_register();
    }
    mutex_unlock(&s_Status.iMutex);
    if(valueHasChanged){
        if(new_value){
            hrtimer_start(&s_Status.iLux_timer, s_Status.iLuxPollDelay,HRTIMER_MODE_REL);
        }else{
            hrtimer_cancel(&s_Status.iLux_timer);
            cancel_work_sync(&s_Status.iLux_work);
        }
    }
    return len;
}//vcnl4020_lux_enable_store
//-----------------------------------------------------------------------------
static ssize_t vcnl4020_lux_input_show(struct device *dev,struct device_attribute *attr, char *buf){
    s_Status.iCountLuxInputShow++;
    return sprintf(buf, "%d\n",s_Status.iLuxValue);
}
//-----------------------------------------------------------------------------
static ssize_t vcnl4020_prox_delay_show(struct device *dev,struct device_attribute *attr, char *buf){
    return sprintf(buf, "%lld\n",ktime_to_ns(s_Status.iProxPollDelay));
}
//-----------------------------------------------------------------------------
static ssize_t vcnl4020_prox_delay_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t len){
    long long val;

    s_Status.iCountProxDelayStore++;
    if (strict_strtoll(buf, 10, &val)){
        return -EINVAL;
    }
    mutex_lock(&s_Status.iMutex);
    s_Status.iRequestedProxPollDelay=val;
    s_Status.iProxPollDelay = get_range_checked_delay(s_Status.iRequestedProxPollDelay);
    mutex_unlock(&s_Status.iMutex);
    return len;
}
//-----------------------------------------------------------------------------
static ssize_t vcnl4020_prox_enable_show(struct device *dev,struct device_attribute *attr, char *buf){
    return sprintf(buf, "%d\n",s_Status.iProxEnabled);
}
//-----------------------------------------------------------------------------
static ssize_t vcnl4020_prox_enable_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t len){
    bool new_value;
    bool valueHasChanged;

    s_Status.iCountProxEnableStore++;
    if (sysfs_streq(buf, "1")){
        new_value = true;
    } else if (sysfs_streq(buf, "0")){
        new_value = false;
    } else {
        return -EINVAL;
    }

    mutex_lock(&s_Status.iMutex);
    valueHasChanged = s_Status.iProxEnabled != new_value;
    if(valueHasChanged){
        s_Status.iProxEnabled = new_value;
        s_Status.iProxValueLastEvent=-1;
        vcnl4020_update_command_register();
    }
    mutex_unlock(&s_Status.iMutex);
    if(valueHasChanged){
        if(new_value){
            hrtimer_start(&s_Status.iProx_timer, s_Status.iProxPollDelay,HRTIMER_MODE_REL);
        }else{
            hrtimer_cancel(&s_Status.iProx_timer);
            cancel_work_sync(&s_Status.iProx_work);
        }
    }
    return len;
}//vcnl4020_prox_enable_store
//-----------------------------------------------------------------------------
static ssize_t vcnl4020_prox_input_show(struct device *dev,struct device_attribute *attr, char *buf){
    s_Status.iCountProxInputShow++;
    return sprintf(buf, "%d\n",s_Status.iProxValue);
}
//-----------------------------------------------------------------------------

static ssize_t vcnl4020_prox_input_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t len){
    if(!sysfs_streq(buf, "0")){
        return -EINVAL;
    }
    mutex_lock(&s_Status.iMutex);
    s_Status.iMaxProx=0;
    s_Status.iMinProx=0;
    memset(s_Status.iProxRingBuf,0,sizeof(s_Status.iProxRingBuf));
    s_Status.iProxMaxThreshold=0;
    s_Status.iProxMinThreshold=0;
    s_Status.iProxValueRound=KValueFar;
    mutex_unlock(&s_Status.iMutex);
    return len;
}//vcnl4020_prox_input_store//*/


static DEVICE_ATTR(lux_enable, S_IRUGO | S_IWUGO, vcnl4020_lux_enable_show, vcnl4020_lux_enable_store);
static DEVICE_ATTR(lux_delay, S_IRUGO | S_IWUGO, vcnl4020_lux_delay_show, vcnl4020_lux_delay_store);
static DEVICE_ATTR(lux_input, S_IRUGO, vcnl4020_lux_input_show, NULL);
static DEVICE_ATTR(prox_enable, S_IRUGO | S_IWUGO, vcnl4020_prox_enable_show, vcnl4020_prox_enable_store);
static DEVICE_ATTR(prox_delay, S_IRUGO | S_IWUGO, vcnl4020_prox_delay_show, vcnl4020_prox_delay_store);
//static DEVICE_ATTR(prox_input, S_IRUGO, vcnl4020_prox_input_show, NULL);
static DEVICE_ATTR(prox_input, S_IRUGO | S_IWUGO, vcnl4020_prox_input_show, vcnl4020_prox_input_store);//nur mal zum test
static DEVICE_ATTR(status, S_IRUGO, vcnl4020_driver_status_show, NULL);

static struct attribute *vcnl4020_sysfs_attrs[] = {
    &dev_attr_lux_delay.attr,
    &dev_attr_lux_enable.attr,
    &dev_attr_lux_input.attr,
    &dev_attr_prox_delay.attr,
    &dev_attr_prox_enable.attr,
    &dev_attr_prox_input.attr,
    &dev_attr_status.attr,
    NULL
};
static struct attribute *vcnl4020_lux_sysfs_attrs[] = {
    &dev_attr_lux_delay.attr,
    &dev_attr_lux_enable.attr,
    &dev_attr_lux_input.attr,
    NULL
};
static struct attribute *vcnl4020_prox_sysfs_attrs[] = {
    &dev_attr_prox_delay.attr,
    &dev_attr_prox_enable.attr,
    &dev_attr_prox_input.attr,
    NULL
};

static struct attribute_group vcnl4020_attribute_group = {
    .name = DRIVER_NAME,
    .attrs = vcnl4020_sysfs_attrs
};

static struct attribute_group vcnl4020_lux_attribute_group = {
    .name = DRIVER_NAME,
    .attrs = vcnl4020_lux_sysfs_attrs
};
static struct attribute_group vcnl4020_prox_attribute_group = {
    .name = DRIVER_NAME,
    .attrs = vcnl4020_prox_sysfs_attrs
};

//-----------------------------------------------------------------------------
static void vcnl4020_lux_work_func(struct work_struct *work){
//    struct MyStatus *pStatus  = container_of(work, struct MyStatus,iLux_work);
    mutex_lock(&s_Status.iMutex);
    if(s_Status.iLuxEnabled && !s_Status.iPowerSuspended){
        //test if data ready
        s_Status.iLuxRequested=true;
        vcnl4020_check_result();
        //report event if changed
        if(s_Status.iLuxValueLastEvent!=s_Status.iLuxValue){
            s_Status.iCountLuxEvent++;
            if(s_Status.iLuxValueLastEvent==-1){
                //it is not possible to send the same value twice (its ignored by the system)
                //so after enable we are sending an invalid value so HAL will get the correct value
                input_report_abs(s_Status.ipLight_input_dev,ABS_MISC, s_Status.iLuxValueLastEvent);
                input_sync(s_Status.ipLight_input_dev);
            }
            s_Status.iLuxValueLastEvent=s_Status.iLuxValue;
            input_report_abs(s_Status.ipLight_input_dev,ABS_MISC, s_Status.iLuxValue);
            input_sync(s_Status.ipLight_input_dev);
        }
    }
    mutex_unlock(&s_Status.iMutex);
}//vcnl4020_lux_work_func

//-----------------------------------------------------------------------------
static enum hrtimer_restart vcnl4020_lux_timer_func(struct hrtimer *timer){
    s_Status.iCountLuxTimer++;
//    struct MyStatus *pStatus  = container_of(timer, struct MyStatus,iLux_timer);

    if(!s_Status.iLuxEnabled || s_Status.iPowerSuspended){
        return HRTIMER_NORESTART;
    }
    queue_work(s_Status.ipWorkQueue, &s_Status.iLux_work);//->vcnl4020_lux_work_func
    hrtimer_forward_now(&s_Status.iLux_timer, s_Status.iLuxPollDelay);
    return HRTIMER_RESTART;
}//vcnl4020_lux_timer_func

//-----------------------------------------------------------------------------
static void vcnl4020_prox_work_func(struct work_struct *work){
//    struct MyStatus *pStatus  = container_of(work, struct MyStatus,iProx_work);
    mutex_lock(&s_Status.iMutex);
    if(s_Status.iProxEnabled && !s_Status.iPowerSuspended){
        //test if data is ready
        s_Status.iProxRequested=true;
        vcnl4020_check_result();
        //report event if changed
        if(s_Status.iProxValueLastEvent!=s_Status.iProxValueRound){
            s_Status.iCountProxEvent++;
            if(s_Status.iProxValueLastEvent==-1){
                //it is not possible to send the same value twice (its ignored by the system)
                //so after enable we are sending an invalid value so HAL will get the correct value
                input_report_abs(s_Status.ipProximity_input_dev,ABS_DISTANCE, s_Status.iProxValueLastEvent);
                input_sync(s_Status.ipProximity_input_dev);
            }
            s_Status.iProxValueLastEvent=s_Status.iProxValueRound;
            input_report_abs(s_Status.ipProximity_input_dev,ABS_DISTANCE, s_Status.iProxValueRound);
            input_sync(s_Status.ipProximity_input_dev);
        }
    }
    mutex_unlock(&s_Status.iMutex);
}//vcnl4020_prox_work_func

//-----------------------------------------------------------------------------
static enum hrtimer_restart vcnl4020_prox_timer_func(struct hrtimer *timer){
    s_Status.iCountProxTimer++;
//    struct MyStatus *pStatus  = container_of(timer, struct MyStatus,iProx_timer);
    if(!s_Status.iProxEnabled || s_Status.iPowerSuspended){
        return HRTIMER_NORESTART;
    }
    queue_work(s_Status.ipWorkQueue, &s_Status.iProx_work);//->vcnl4020_prox_work_func
    hrtimer_forward_now(&s_Status.iProx_timer, s_Status.iProxPollDelay);
    return HRTIMER_RESTART;
}

//-----------------------------------------------------------------------------
//will be called from vcnl4020_probe() is only b
static int vcnl4020_set_default_config(struct i2c_client *client){
    int ret;
    //disable measurment
    ret = i2c_smbus_write_byte_data(client, VCNL4020_COMMAND_REGISTER, VCNL4020_COMMAND_REGISTER__ALL_DISABLE);
    s_Status.iLastCommandRegisterWrite=VCNL4020_COMMAND_REGISTER__ALL_DISABLE;
    if (ret < 0) {
        dev_err(&client->dev, "%s: VCNL4020_COMMAND_REGISTER__ALL_DISABLE err=%d\n",__func__,ret);
        return ret;
    }
    //VCNL4020_PROXIMITY_RATE_REGISTER
    //0=1.95 1=3.90625 2=7.8125 3=15,625 4=31,25 5=62,5 6=125 7=250  measurments/s
    //0=512 1=256 2=128 3=64 4=32 5=16 6=8 7=4  time in ms
    ret = i2c_smbus_write_byte_data(client, VCNL4020_PROXIMITY_RATE_REGISTER, 2);
    if (ret < 0) {
        dev_err(&client->dev, "%s: VCNL4020_PROXIMITY_RATE_REGISTER err=%d\n",__func__,ret);
        return ret;
    }
    //VCNL4020_IR_LED_CURRENT_REGISTER
    //0=0 1=20 ... 20=max=200 mA
    ret = i2c_smbus_write_byte_data(client, VCNL4020_IR_LED_CURRENT_REGISTER, 20);
    if (ret < 0) {
        dev_err(&client->dev, "%s: VCNL4020_IR_LED_CURRENT_REGISTER err=%d\n",__func__,ret);
        return ret;
    }
    //VCNL4020_AMBIENT_LIGHT_PARAMETER_REGISTER
    //bit_7_ContMode=0
    //bit_6_4_als_rate 0=1 1=2 3=4 4=5 5=6 6=8 7=10 measurments/s
    //                 0=1000 1=500 2=333 3=250 4=200 5=167 6=125 7=100 time in ms
    //bit_3_auto_offset
    //bit 2_0_Averaging 2^n conversion/measurment 0=1 1=2 2=4 3=8 4=16 5=32 6=64 7=128
    ret = i2c_smbus_write_byte_data(client, VCNL4020_AMBIENT_LIGHT_PARAMETER_REGISTER, ((5<<4)|(1<<3)|(7)));
    if (ret < 0) {
        dev_err(&client->dev, "%s: VCNL4020_AMBIENT_LIGHT_PARAMETER_REGISTER err=%d\n",__func__,ret);
        return ret;
    }
    //VCNL4020_INTERRUPT_CONTROL_REGISTER
    ret = i2c_smbus_write_byte_data(client, VCNL4020_INTERRUPT_CONTROL_REGISTER, 0);
    if (ret < 0) {
        dev_err(&client->dev, "%s: VCNL4020_INTERRUPT_CONTROL_REGISTER err=%d\n",__func__,ret);
        return ret;
    }

    return 0;
}//vcnl4020_set_default_config

//-----------------------------------------------------------------------------
static int vcnl4020_probe(struct i2c_client *client,const struct i2c_device_id *id){	
    int res = -ENODEV;

    s_Status.iCountProbe++;

    if (!i2c_check_functionality(client->adapter
                                , I2C_FUNC_I2C
                                 |I2C_FUNC_SMBUS_READ_BYTE_DATA
                                 |I2C_FUNC_SMBUS_READ_BYTE
                                 |I2C_FUNC_SMBUS_WRITE_BYTE_DATA
                                )) {
        dev_err(&client->dev, "%s: i2c_check_functionality failed\n", __func__);
        return -ENODEV;
    }

    //i2c_set_clientdata(client, pStatus);
    i2c_set_clientdata(client, &s_Status);
    s_Status.ip_i2c_client = client;

    //test if we can find our sensor chip on board
    res = i2c_smbus_read_byte_data(client, VCNL4020_PRODUCT_ID_REVISION_REGISTER);
    if( res < 0 ) {
        dev_err(&client->dev, "read VCNL4020_PRODUCT_ID_REVISION_REGISTER err=%d\n",res);
        return res;
    }
    if( res != 0x21 ) {
        dev_err(&client->dev, "read VCNL4020_PRODUCT_ID_REVISION_REGISTER unknown version=0x%x\n",res);
        return -ENODEV;
    }

    //ToDo brauchen wir auch einen wake lock??        wake_lock_init(&cm36651->prx_wake_lock, WAKE_LOCK_SUSPEND,"prx_wake_lock");
    mutex_init(&s_Status.iMutex);

    //setup initial chip registers
    res = vcnl4020_set_default_config(client);
    if (res <  0){
        dev_err(&client->dev, "%s: err_set_default_config e=%d\n",__func__,res);
        goto err_set_default_config;
    }

    //request_threaded_irq machen wir nicht da wir den IRQ erst mal nicht benutzen

    //initialize proximity input_device
    s_Status.ipProximity_input_dev = input_allocate_device();
    if (!s_Status.ipProximity_input_dev) {
        dev_err(&client->dev, "%s: err_prox_input_allocate_device e=%d\n",__func__,res);
        goto err_prox_input_allocate_device;
    }
    input_set_drvdata(s_Status.ipProximity_input_dev, &s_Status);
    s_Status.ipProximity_input_dev->name = "proximity_sensor";
    input_set_capability(s_Status.ipProximity_input_dev, EV_ABS,ABS_DISTANCE);
    input_set_abs_params(s_Status.ipProximity_input_dev, ABS_DISTANCE// unsigned int axis,
                                                       , 0//int min
                                                       , 100//, int max
                                                       , 0, 0);//, int fuzz, int flat);
    res = input_register_device(s_Status.ipProximity_input_dev);
    if (res < 0) {
        dev_err(&client->dev, "%s: err_prox_input_register_device e=%d\n",__func__,res);
        goto err_prox_input_register_device;
    }
    res = sysfs_create_group(&s_Status.ipProximity_input_dev->dev.kobj,&vcnl4020_prox_attribute_group);
    if (res) {
        dev_err(&client->dev, "%s: err_prox_sysfs_create_group e=%d\n",__func__,res);
        goto err_prox_sysfs_create_group;
    }

    //initialize light input_device
    s_Status.ipLight_input_dev = input_allocate_device();
    if (!s_Status.ipLight_input_dev) {
        dev_err(&client->dev, "%s: err_lux_input_allocate_device e=%d\n",__func__,res);
        goto err_lux_input_allocate_device;
    }
    input_set_drvdata(s_Status.ipLight_input_dev, &s_Status);
    s_Status.ipLight_input_dev->name = "light_sensor";
    input_set_capability(s_Status.ipLight_input_dev, EV_ABS,ABS_MISC);
    input_set_abs_params(s_Status.ipLight_input_dev, ABS_MISC// unsigned int axis,
                                                   , 0//int min
                                                   , 1000//, int max
                                                   , 0, 0);//, int fuzz, int flat);
    res = input_register_device(s_Status.ipLight_input_dev);
    if (res < 0) {
        dev_err(&client->dev, "%s: err_lux_input_register_device e=%d\n",__func__,res);
        goto err_lux_input_register_device;
    }
    res = sysfs_create_group(&s_Status.ipLight_input_dev->dev.kobj,&vcnl4020_lux_attribute_group);
    if (res) {
        dev_err(&client->dev, "%s: err_lux_sysfs_create_group e=%d\n",__func__,res);
        goto err_lux_sysfs_create_group;
    }


    //create file in sysfs to communicate with user space / android HAL Layer
    res = sysfs_create_group(&client->dev.kobj, &vcnl4020_attribute_group);
    if (res) {
        dev_err(&client->dev, "%s: err_sysfs_create_group e=%d\n",__func__,res);
        goto err_sysfs_create_group;
    }
    dev_info(&client->dev, "%s chip found\n", client->name);

    //initialize timer and workqueue for polling
    //we need the workqueue as i2c must run inside timer call back 
    s_Status.ipWorkQueue = create_singlethread_workqueue("vcnl4020_worker");
    if (!s_Status.ipWorkQueue) {
        res = -ENOMEM;
        dev_err(&client->dev, "%s: err_create_workqueue e=%d\n",__func__,res);
        goto err_create_workqueue;
    }

    hrtimer_init(&s_Status.iProx_timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
    s_Status.iProxPollDelay = ns_to_ktime(199 * NSEC_PER_MSEC);
    s_Status.iProx_timer.function = vcnl4020_prox_timer_func;
    INIT_WORK(&s_Status.iProx_work, vcnl4020_prox_work_func);

    hrtimer_init(&s_Status.iLux_timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
    s_Status.iLuxPollDelay = ns_to_ktime(199 * NSEC_PER_MSEC);
    s_Status.iLux_timer.function = vcnl4020_lux_timer_func;
    INIT_WORK(&s_Status.iLux_work, vcnl4020_lux_work_func);

    res=0;
    goto done;
    //error handling -> unwind resources!!
    destroy_workqueue(s_Status.ipWorkQueue);
err_create_workqueue:
    sysfs_remove_group(&client->dev.kobj, &vcnl4020_attribute_group);
err_sysfs_create_group:
    sysfs_remove_group(&s_Status.ipLight_input_dev->dev.kobj,&vcnl4020_lux_attribute_group);
err_lux_sysfs_create_group:
    input_unregister_device(s_Status.ipLight_input_dev);
    s_Status.ipLight_input_dev=0;
err_lux_input_register_device:
    if(s_Status.ipLight_input_dev){
        input_free_device(s_Status.ipLight_input_dev);
        s_Status.ipLight_input_dev=0;
    }
err_lux_input_allocate_device:
    sysfs_remove_group(&s_Status.ipProximity_input_dev->dev.kobj,&vcnl4020_prox_attribute_group);
err_prox_sysfs_create_group:
    input_unregister_device(s_Status.ipProximity_input_dev);
    s_Status.ipProximity_input_dev=0;
err_prox_input_register_device:
    if(s_Status.ipProximity_input_dev){
        input_free_device(s_Status.ipProximity_input_dev);
        s_Status.ipProximity_input_dev=0;
    }
err_prox_input_allocate_device:
err_set_default_config:
    mutex_destroy(&s_Status.iMutex);
done:
    return res;
}//vcnl4020_probe

//-----------------------------------------------------------------------------
//remove all resources used by our driver
static int vcnl4020_remove(struct i2c_client *client){
    s_Status.iCountRemove++;
//cancel timer
//stop sensor
    input_unregister_device(s_Status.ipProximity_input_dev);
    destroy_workqueue(s_Status.ipWorkQueue);
    sysfs_remove_group(&s_Status.ipLight_input_dev->dev.kobj,&vcnl4020_lux_attribute_group);
    sysfs_remove_group(&s_Status.ipProximity_input_dev->dev.kobj,&vcnl4020_prox_attribute_group);
    sysfs_remove_group(&client->dev.kobj, &vcnl4020_attribute_group);
    mutex_destroy(&s_Status.iMutex);
    return 0;
}//vcnl4020_remove

//-----------------------------------------------------------------------------
//power managment
static int vcnl4020_suspend(struct device *dev){
//    struct i2c_client *client = to_i2c_client(dev);
    s_Status.iCountSuspend++;
    mutex_lock(&s_Status.iMutex);
    s_Status.iPowerSuspended=true;
    vcnl4020_update_command_register();
    mutex_unlock(&s_Status.iMutex);

    if(s_Status.iLuxEnabled){
        hrtimer_cancel(&s_Status.iLux_timer);
        cancel_work_sync(&s_Status.iLux_work);
    }
    if(s_Status.iProxEnabled){
        hrtimer_cancel(&s_Status.iProx_timer);
        cancel_work_sync(&s_Status.iProx_work);
    }

    return 0;
}//vcnl4020_suspend

//-----------------------------------------------------------------------------
//power managment
static int vcnl4020_resume(struct device *dev){
//    struct i2c_client *client = to_i2c_client(dev);
    s_Status.iCountResume++;
    mutex_lock(&s_Status.iMutex);
    s_Status.iPowerSuspended=false;
    vcnl4020_update_command_register();
    mutex_unlock(&s_Status.iMutex);
    if(s_Status.iLuxEnabled){
        hrtimer_start(&s_Status.iLux_timer, s_Status.iLuxPollDelay,HRTIMER_MODE_REL);
    }
    if(s_Status.iProxEnabled){
        hrtimer_start(&s_Status.iProx_timer, s_Status.iProxPollDelay,HRTIMER_MODE_REL);
    }
    return 0;
}//vcnl4020_resume

//#############################################################################

static struct i2c_device_id vcnl4020_id[] = {
    { DRIVER_NAME, 0 },
    {}
};
MODULE_DEVICE_TABLE(i2c, vcnl4020_id);

static const struct dev_pm_ops vcnl4020_pm_ops = {
    .suspend = vcnl4020_suspend,
    .resume = vcnl4020_resume
};

static struct i2c_driver vcnl4020_driver = {
    .driver = {
        .name = DRIVER_NAME,
        .owner = THIS_MODULE,
        .pm = &vcnl4020_pm_ops
    },
    .id_table = vcnl4020_id,
    .probe = vcnl4020_probe,
    .remove = vcnl4020_remove,
};

//-----------------------------------------------------------------------------
static int __init sensor_vcnl4020_init(void){
    int res;

    s_Status.iCountInit++;
    res = i2c_add_driver(&vcnl4020_driver);
    return res;
}

//-----------------------------------------------------------------------------
static void  __exit sensor_vcnl4020_exit(void){
    s_Status.iCountExit++;
    i2c_del_driver(&vcnl4020_driver);
}

module_init(sensor_vcnl4020_init);
module_exit(sensor_vcnl4020_exit);

MODULE_AUTHOR("Volker Hoppe <volker.hoppe@sic-software.com>");
MODULE_DESCRIPTION("VISHAY VCNL4020 driver");
MODULE_LICENSE("GPL v2");

